;; Time-stamp: <> -*-nasm-*-
;; toupper.asm: This program converts a file to to uppercase.
 
; Processing: 1) Open the input file 
;             2) Open the output file 
;             4) While we're not at the end of the input file 
;                a) read part of file into our memory buffer 
;                b) go through each byte of memory 
;                     if the byte is a lower-case letter, 
;                     convert it to uppercase 
;                c) write the memory buffer to output file 

          section   .data 
;;;;;;;CONSTANTS;;;;;;;; 

; system call numbers 

SYS_OPEN  equ       5 
SYS_WRITE equ       4 
SYS_READ  equ       3 
SYS_CLOSE equ       6 
SYS_EXIT  equ       1 

; Options for open (look at /usr/include/asm/fcntl.h for various
; values.  You can combine them by adding them or ORing them).  This
; is discussed at greater length in "Counting Like a Computer"

O_RDONLY             equ 0 
O_CREAT_WRONLY_TRUNC equ 3101q 

; standard file descriptors 

STDIN     equ       0 
STDOUT    equ       1 
STDERR    equ       2 

; system call interrupt 

LINUX_SYSCALL equ   80h    
END_OF_FILE   equ   0             ; This is the return value 
                                  ; of read which means we've 
                                  ; hit the end of the file 
NUMBER_ARGUMENTS equ 2 

;                   section   .bss 
; Buffer - this is where the data is loaded into from the data file
;  and written from into the output file.  This should never exceed
;  16,000 for various reasons.

BUFFER_SIZE equ     500 
BUFFER_DATA         times     BUFFER_SIZE db 0

          global    _start 

          section   .text 
; STACK POSITIONS 
ST_SIZE_RESERVE equ 8 
ST_FD_IN  equ       -4 
ST_FD_OUT equ       -8 
ST_ARGC   equ       0             ; Number of arguments 
ST_ARGV_0 equ       4             ; Name of program 
ST_ARGV_1 equ       8             ; Input file name 
ST_ARGV_2 equ       12            ; Output file name 

;;;INITIALIZE PROGRAM;;; 

; save the stack pointer 

_start:   mov       ebp,esp 

; Allocate space for our file descriptors on the stack

          sub       esp,ST_SIZE_RESERVE 
open_files: 
open_fd_in: 

;;;OPEN INPUT FILE;;; 

; open syscall 

          mov       eax,SYS_OPEN 

; input filename into %ebx 

          mov       ebx,[ebp+ST_ARGV_1] 

; read-only flag 

          mov       ecx,O_RDONLY 

; this doesn't really matter for reading 

          mov       edx,666q 

; call Linux 

          int       LINUX_SYSCALL 
store_fd_in: 

; save the given file descriptor 

          mov       [ebp+ST_FD_IN],eax 
open_fd_out: 

;;;OPEN OUTPUT FILE;;; 

; open the file 

          mov       eax,SYS_OPEN 

; output filename into EBX 
 
          mov       ebx,[ebp+ST_ARGV_2] 

; flags for writing to the file 

          mov       ecx,O_CREAT_WRONLY_TRUNC 

; mode for new file (if it's created) 

          mov       edx,666q 

; call Linux 

          int       LINUX_SYSCALL 
store_fd_out: 

; store the file descriptor here 

          mov       [ebp+ST_FD_OUT],eax 

;;;BEGIN MAIN LOOP;;; 

read_loop_begin: 

;;;READ IN A BLOCK FROM THE INPUT FILE;;; 

          mov       eax,SYS_READ 

; get the input file descriptor 

          mov       ebx,[ebp+ST_FD_IN] 

; the location to read into 

          mov       ecx,BUFFER_DATA 

; the size of the buffer 

          mov       edx,BUFFER_SIZE 

; Size of buffer read is returned in %eax 

          int       LINUX_SYSCALL 

;;;EXIT IF WE'VE REACHED THE END;;; 

; check for end of file marker 

          cmp       eax,END_OF_FILE 

; if found or on error, go to the end 

          jle       end_loop 
continue_read_loop: 

;;;CONVERT THE BLOCK TO UPPER CASE;;; 

          push      BUFFER_DATA   ; location of buffer 
          push      eax           ; size of the buffer 
          call      convert_to_upper 
          pop       eax           ; get the size back 
          add       esp,4         ; restore %esp 

;;;WRITE THE BLOCK OUT TO THE OUTPUT FILE;;; 

; size of the buffer 

          mov       edx,eax 
          mov       eax,SYS_WRITE 

; file to use 

          mov       ebx,[ebp+ST_FD_OUT] 

; location of the buffer 

          mov       ecx,BUFFER_DATA 
          int       LINUX_SYSCALL 

;;;CONTINUE THE LOOP;;; 

          jmp       read_loop_begin 
end_loop: 

;;;CLOSE THE FILES;;; 

; NOTE - we don't need to do error checking on these, because error
;  conditions don't signify anything special here

          mov       eax,SYS_CLOSE 
          mov       ebx,[ebp+ST_FD_OUT] 
          int       LINUX_SYSCALL 
          mov       eax,SYS_CLOSE 
          mov       ebx,[ebp+ST_FD_IN] 
          int       LINUX_SYSCALL 
;;;EXIT;;; 

          mov       eax,SYS_EXIT 
          mov       ebx,0 
          int       LINUX_SYSCALL 

; Purpose: This function actually does the conversion to upper case
;          for a block
 
; Entry conditions: The first parameter is the location of the block
;                   of memory to convert.  The second parameter is the
;                   length of that buffer
 
; Exit conditions: This function overwrites the current buffer with
;                  the upper-casified version.
 
; Variables: 
;                  EAX - beginning of buffer 
;                  EBX - length of buffer 
;                  EDI - current buffer offset 
;                  CL - current byte being examined (first part of
;                  ECX)
 
;;;CONSTANTS;; 

; The lower boundary of our search 

LOWERCASE_A equ     'a' 

; The upper boundary of our search 

LOWERCASE_Z equ     'z' 

; Conversion between upper and lower case 

UPPER_CONVERSION equ          'A' - 'a' 

;;;STACK STUFF;;; 

ST_BUFFER_LEN equ   8             ; Length of buffer 
ST_BUFFER     equ   12            ; actual buffer 

convert_to_upper: 
          push      ebp 
                    mov       ebp,esp 

;;;SET UP VARIABLES;;; 

          mov       eax,[ebp+ST_BUFFER] 
          mov       ebx,[ebp+ST_BUFFER_LEN] 
          mov       edi,0 

; if a buffer with zero length was given to us, just leave

          cmp       ebx,0 
          je        end_convert_loop 
convert_loop: 

; get the current byte 

          mov       cl,[edi+eax+1] 

; go to the next byte unless it is between 'a' and 'z'

          cmp       cl,LOWERCASE_A 
          jl        next_byte 
          cmp       cl,LOWERCASE_Z 
          jg        next_byte 

; otherwise convert the byte to uppercase 

          add       cl,UPPER_CONVERSION 

; and store it back 

          mov       [edi+eax+1],cl 
next_byte: 
          inc       edi           ; next byte 
          cmp       ebx,edi       ; continue unless 

; we've reached the end

          jne       convert_loop 
end_convert_loop: 

; no return value, just leave 

          mov       esp,ebp 
          pop       ebp 
          ret       
